@@ -41,6 +41,7 @@
 #include <gio/gio.h>
 #include <glib/gprintf.h>
 
+#include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <locale.h>
@@ -48,6 +49,7 @@
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/statvfs.h>
 #include <sys/wait.h>
 #include <unistd.h> /* unlinkat, fork */
 
@@ -60,6 +62,57 @@ typedef struct
   int dummy;
 } TestDummy;
 
+/* test tables for string to number conversions */
+struct
+{
+  const char *str;
+  gint        val;
+  gboolean    error;
+} str_to_int_table[] = {
+  {"0",                                        0,                FALSE},
+  {"1",                                        1,                FALSE},
+  {"-1",                                      -1,                FALSE},
+#if __SIZEOF_INT__ == 4
+  {"2147483647",                      2147483647,                FALSE}, /* MAX_INT */
+  {"-2147483648",                    -2147483648,                FALSE}, /* MIN_INT */
+  {"2147483648",                               0,                TRUE},  /* MAX_INT + 1 */
+  {"-2147483649",                              0,                TRUE},  /* MIN_INT - 1 */
+#elif __SIZEOF_INT__ == 8
+  {"9223372036854775807",    9223372036854775807,                FALSE}, /* MAX_INT */
+  {"-9223372036854775808",  -9223372036854775808,                FALSE}, /* MIN_INT */
+  {"9223372036854775808",                      0,                TRUE},  /* MAX_INT + 1 */
+  {"-9223372036854775809",                     0,                TRUE},  /* MIN_INT - 1 */
+#else
+#warning __SIZEOF_INT__ not handled
+#endif
+  {"notanint",                                 0,                TRUE},
+  {"9223372036854775808",                      0,                TRUE}, /* overflow */
+  {"-9223372036854775809",                     0,                TRUE}, /* underflow */
+};
+
+struct
+{
+  const char *str;
+  guint       val;
+  gboolean    error;
+} str_to_uint_table[] = {
+  {"0",                                        0,                FALSE},
+  {"1",                                        1,                FALSE},
+  {"-1",                                       0,                TRUE}, /* negative */
+#if __SIZEOF_INT__ == 4
+  {"4294967295",                      4294967295,                FALSE}, /* MAX_UINT */
+  {"4294967296",                               0,                TRUE},  /* MAX_UINT + 1 */
+#elif __SIZEOF_INT__ == 8
+  {"18446744073709551615",  18446744073709551615,                FALSE}, /* MAX_INT */
+  {"18446744073709551616",                     0,                TRUE},  /* MAX_INT + 1 */
+#else
+#warning __SIZEOF_INT__ not handled
+#endif
+  {"notanint",                                 0,                TRUE},
+  {"18446744073709551617",                     0,                TRUE}, /* overflow */
+};
+
+
 #define TEST_DBUS_GRESOURCE_PATH "/bolt/tests/exported/example.bolt.xml"
 #define TEST_DBUS_INTERFACE "org.gnome.bolt.Example"
 
@@ -126,7 +179,7 @@ typedef struct
 static void
 test_enums (TestRng *tt, gconstpointer user_data)
 {
-  g_autoptr(GEnumClass) klass;
+  g_autoptr(GEnumClass) klass = NULL;
   g_autoptr(GError) err = NULL;
   const char *str;
   gint val;
@@ -362,7 +415,7 @@ test_error (TestRng *tt, gconstpointer user_data)
 static void
 test_flags (TestRng *tt, gconstpointer user_data)
 {
-  g_autoptr(GFlagsClass) klass;
+  g_autoptr(GFlagsClass) klass = NULL;
   g_autoptr(GError) err = NULL;
   char *str;
   guint val;
@@ -672,6 +725,7 @@ test_io_errors (TestIO *tt, gconstpointer user_data)
   struct stat st;
   char buffer[256] = {0, };
   gboolean ok;
+  unsigned int uiv;
   int fd = -1;
   int iv;
   int r;
@@ -828,6 +882,17 @@ test_io_errors (TestIO *tt, gconstpointer user_data)
   g_assert_false (ok);
   g_clear_pointer (&err, g_error_free);
 
+  /* read_uint_at */
+  ok = bolt_read_uint_at (dirfd (root), "NONEXISTENT", &uiv, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_FOUND);
+  g_assert_false (ok);
+  g_clear_pointer (&err, g_error_free);
+
+  ok = bolt_read_uint_at (dirfd (root), "readonly", &uiv, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
+  g_assert_false (ok);
+  g_clear_pointer (&err, g_error_free);
+
   /* pipe error checking */
   fd = bolt_mkfifo (tt->path, 0600, &err);
   g_assert_error (err, G_IO_ERROR, G_IO_ERROR_EXISTS);
@@ -878,6 +943,34 @@ test_io_errors (TestIO *tt, gconstpointer user_data)
   g_clear_pointer (&err, g_error_free);
 }
 
+static void
+test_io_mkdirat (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) d = NULL;
+  struct stat st;
+  gboolean ok;
+
+  d = bolt_opendir (tt->path, &error);
+
+  g_assert_no_error (error);
+  g_assert_nonnull (d);
+
+  ok = bolt_mkdirat (dirfd (d), "directory", 0666, &error);
+  g_assert_no_error (error);
+  g_assert_true (ok);
+
+  ok = bolt_fstatat (dirfd (d), "directory", &st, 0, &error);
+  g_assert_no_error (error);
+  g_assert_true (ok);
+
+  g_assert_true (S_ISDIR (st.st_mode));
+
+  ok = bolt_mkdirat (dirfd (d), "directory", 0666, &error);
+  g_assert_error (error, G_IO_ERROR, G_IO_ERROR_EXISTS);
+  g_assert_false (ok);
+}
+
 static void
 test_io_verify (TestIO *tt, gconstpointer user_data)
 {
@@ -932,13 +1025,186 @@ test_io_verify (TestIO *tt, gconstpointer user_data)
   unlinkat (dirfd (d), "unique_id", 0);
 }
 
+static void
+test_io_write_file_at (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) dir = NULL;
+  g_autofree char *path = NULL;
+  g_autofree char *data = NULL;
+  static const char *ref = "The world is everything that is the case.";
+  gboolean ok;
+  gsize len;
+
+  dir = bolt_opendir (tt->path, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (dir);
+
+  ok = bolt_write_file_at (dirfd (dir), "test.txt", ref, -1, &error);
+
+  g_assert_no_error (error);
+  g_assert_true (ok);
+
+  path = g_build_filename (tt->path, "test.txt", NULL);
+  ok = g_file_get_contents (path, &data, &len, &error);
+  g_assert_no_error (error);
+  g_assert_true (ok);
+  g_assert_cmpuint (strlen (ref), ==, len);
+  g_assert_cmpstr (ref, ==, data);
+
+  g_clear_pointer (&data, g_free);
+  ok = bolt_file_write_all (path, ref, 5, &error);
+
+  g_assert_no_error (error);
+  g_assert_true (ok);
+
+  ok = g_file_get_contents (path, &data, &len, &error);
+  g_assert_no_error (error);
+  g_assert_true (ok);
+  g_assert_cmpuint (len, ==, 5);
+  g_assert_true (strncmp (data, ref, 5) == 0);
+}
+
+static void
+test_io_write_int_at (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) dir = NULL;
+  gboolean ok;
+  int tests[] = {0, 1, 42, G_MAXINT};
+
+  dir = bolt_opendir (tt->path, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (dir);
+
+  for (gsize i = 0; i < G_N_ELEMENTS (tests); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      int ref = tests[i];
+      int val;
+
+      ok = bolt_write_int_at (dirfd (dir), "int.txt", ref, &err);
+      g_assert_true (ok);
+
+      ok = bolt_read_int_at (dirfd (dir), "int.txt", &val, &err);
+      g_assert_no_error (err);
+      g_assert_true (ok);
+      g_assert_cmpint (val, ==, ref);
+    }
+}
+
+static void
+test_io_read_int_at (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) dir = NULL;
+  gboolean ok;
+
+  dir = bolt_opendir (tt->path, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (dir);
+
+  for (gsize i = 0; i < G_N_ELEMENTS (str_to_int_table); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      const char *txt = str_to_int_table[i].str;
+      gboolean expect_error = str_to_int_table[i].error;
+      gint val = str_to_int_table[i].val;
+      gint v;
+
+      ok = bolt_write_file_at (dirfd (dir), "int.txt", txt, -1, &err);
+      g_assert_true (ok);
+
+      ok = bolt_read_int_at (dirfd (dir), "int.txt", &v, &err);
+      if (expect_error)
+        {
+          g_assert_nonnull (err);
+          g_assert_false (ok);
+        }
+      else
+        {
+          g_assert_no_error (err);
+          g_assert_cmpint (val, ==, v);
+          g_assert_true (ok);
+        }
+    }
+}
+
+static void
+test_io_write_uint_at (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) dir = NULL;
+  gboolean ok;
+  unsigned int tests[] = {42, 0, 1, G_MAXUINT};
+
+  dir = bolt_opendir (tt->path, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (dir);
+
+  for (gsize i = 0; i < G_N_ELEMENTS (tests); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      unsigned int ref = tests[i];
+      unsigned int val;
+
+      ok = bolt_write_uint_at (dirfd (dir), "uint.txt", ref, &err);
+      g_assert_true (ok);
+
+      ok = bolt_read_uint_at (dirfd (dir), "uint.txt", &val, &err);
+      g_assert_no_error (err);
+      g_assert_true (ok);
+      g_assert_cmpint (val, ==, ref);
+    }
+}
+
+static void
+test_io_read_uint_at (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(DIR) dir = NULL;
+  gboolean ok;
+
+  dir = bolt_opendir (tt->path, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (dir);
+
+  for (gsize i = 0; i < G_N_ELEMENTS (str_to_uint_table); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      const char *txt = str_to_uint_table[i].str;
+      gboolean expect_error = str_to_uint_table[i].error;
+      guint val = str_to_uint_table[i].val;
+      guint v;
+
+      if (g_test_verbose ())
+        g_test_message ("bolt_read_uint: '%s'", txt);
+
+      ok = bolt_write_file_at (dirfd (dir), "uint.txt", txt, -1, &err);
+      g_assert_true (ok);
+
+      ok = bolt_read_uint_at (dirfd (dir), "uint.txt", &v, &err);
+      if (expect_error)
+        {
+          g_assert_nonnull (err);
+          g_assert_false (ok);
+        }
+      else
+        {
+          g_assert_no_error (err);
+          g_assert_cmpuint (val, ==, v);
+          g_assert_true (ok);
+        }
+    }
+}
+
 static void
 test_io_file_write_all (TestIO *tt, gconstpointer user_data)
 {
   g_autoptr(GError) error = NULL;
   g_autofree char *path = NULL;
   g_autofree char *data = NULL;
-  static const char *ref = "Die Welt is alles was der Fall ist!";
+  static const char *ref = "The world is everything that is the case.";
   gboolean ok;
   gsize len;
 
@@ -967,6 +1233,67 @@ test_io_file_write_all (TestIO *tt, gconstpointer user_data)
   g_assert_true (strncmp (data, ref, 5) == 0);
 }
 
+static void
+test_io_renameat (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(DIR) root = NULL;
+  g_autoptr(DIR) subdir = NULL;
+  struct stat st;
+  gboolean ok;
+
+  root = bolt_opendir (tt->path, &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (root);
+
+  ok = bolt_write_file_at (dirfd (root), "a", "a", -1, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  ok = bolt_mkdirat (dirfd (root), "subdir", 0777, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  subdir = bolt_opendir_at (dirfd (root), "subdir", O_RDONLY, &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (subdir);
+
+  ok = bolt_renameat (dirfd (root), "a",
+                      dirfd (subdir), "b",
+                      &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  ok = bolt_fstatat (dirfd (subdir), "b", &st, 0, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  g_assert_true (S_ISREG (st.st_mode));
+
+  ok = bolt_renameat (dirfd (subdir), "b",
+                      dirfd (subdir), "c",
+                      &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  memset (&st, 0, sizeof (st));
+  ok = bolt_fstatat (dirfd (subdir), "c", &st, 0, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  g_assert_true (S_ISREG (st.st_mode));
+
+  /* error reporting: file not found  */
+  ok = bolt_renameat (dirfd (subdir), "b",
+                      dirfd (subdir), "c",
+                      &err);
+
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_FOUND);
+  g_assert_false (ok);
+}
+
 static void
 test_io_copy_bytes (TestIO *tt, gconstpointer user_data)
 {
@@ -981,6 +1308,11 @@ test_io_copy_bytes (TestIO *tt, gconstpointer user_data)
   int from = -1;
   int to = -1;
 
+  /* bolt_copy_bytes uses copy_file_range(2) internally, which
+   * was added in linux 4.5. */
+  skip_test_unless (bolt_check_kernel_version (4, 5) || g_test_thorough (),
+                    "linux kernel < 4.5, copy_file_range syscall missing");
+
   chk = g_checksum_new (G_CHECKSUM_SHA256);
 
   source = g_build_filename (tt->path, "copy_bytes_source", NULL);
@@ -1051,6 +1383,72 @@ test_io_copy_bytes (TestIO *tt, gconstpointer user_data)
                    chksum);
 }
 
+static void
+test_io_dir_is_empty (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(DIR) root = NULL;
+  struct dirent *de = NULL;
+  gboolean empty;
+  gboolean ok;
+  long pos;
+
+  root = bolt_opendir (tt->path, &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (root);
+
+  empty = FALSE;
+  ok = bolt_dir_is_empty (root, &empty, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_true (empty);
+
+  ok = bolt_write_file_at (dirfd (root), "a", "a", -1, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  empty = TRUE;
+  ok = bolt_dir_is_empty (root, &empty, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_false (empty);
+
+  /* check that we don't mess with the dir pointer offset,
+   * by creating a second file, iterating exactly once
+   * and then making sure we are at the same position
+   * after the call to bolt_dir_is_empty */
+
+  ok = bolt_write_file_at (dirfd (root), "b", "b", -1, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  empty = TRUE;
+  ok = bolt_dir_is_empty (root, &empty, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_false (empty);
+
+  while ((de = readdir (root)) != NULL)
+    {
+
+      if (!g_strcmp0 (de->d_name, ".") ||
+          !g_strcmp0 (de->d_name, ".."))
+        continue;
+
+      break;
+    }
+
+  pos = telldir (root);
+
+  empty = TRUE;
+  ok = bolt_dir_is_empty (root, &empty, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_false (empty);
+
+  g_assert_cmpint (telldir (root), ==, pos);
+}
+
 static void
 test_autoclose (TestIO *tt, gconstpointer user_data)
 {
@@ -1127,7 +1525,7 @@ test_fs (TestIO *tt, gconstpointer user_data)
 
   if (geteuid () == 0)
     /* if we run as root, maybe inside a container, we will
-     *  be able o do that anyway so skip it in that case */
+     * be able to do that anyway so skip it in that case */
     return;
 
   /* check error checking */
@@ -1170,13 +1568,13 @@ touch_and_compare (GFile *target, guint64 tp)
   ts = g_file_info_get_attribute_uint64 (info, "time::modified");
   g_assert_cmpuint (ts, ==, tp + 1);
 
-  ts = g_file_info_get_attribute_uint64 (info, "time::access");
-  g_assert_cmpuint (ts, ==, tp);
+  /* No check for access time because it might not be supported
+   * by the underlying file system */
 }
 
 
 static void
-test_fs_touch (TestIO *tt, gconstpointer user_data)
+test_fs_touch_mtime (TestIO *tt, gconstpointer user_data)
 {
   g_autoptr(GFileInfo) info = NULL;
   g_autoptr(GError) err = NULL;
@@ -1199,7 +1597,7 @@ test_fs_touch (TestIO *tt, gconstpointer user_data)
   tp = 626648700;
   touch_and_compare (target, tp);
 
-  /* omit one of them, start with atime */
+  /* check mtime, omit atime */
   ok = bolt_fs_touch (target, 0, 42, &err);
   g_assert_no_error (err);
   g_assert_true (ok);
@@ -1209,13 +1607,72 @@ test_fs_touch (TestIO *tt, gconstpointer user_data)
   g_assert_no_error (err);
   g_assert_nonnull (info);
 
+  ts = g_file_info_get_attribute_uint64 (info, "time::modified");
+  g_assert_cmpuint (ts, ==, 42);
+
+  /* the other way around, where 0 means now */
+  ok = bolt_fs_touch (target, 42, 0, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  g_clear_object (&info);
+  info = g_file_query_info (target, TIME_QI, 0, NULL, &err);
+
+  g_assert_no_error (err);
+  g_assert_nonnull (info);
+
+  /* mtime 0 means ignore, effectively meaning now, for touch */
+  ts = g_file_info_get_attribute_uint64 (info, "time::modified");
+  g_assert_cmpuint (ts, >=, now);
+  now = (guint64) g_get_real_time () / G_USEC_PER_SEC;
+  g_assert_cmpuint (ts, <=, now);
+}
+
+static void
+test_fs_touch_atime (TestIO *tt, gconstpointer user_data)
+{
+  g_autoptr(GFileInfo) info = NULL;
+  g_autoptr(GError) err = NULL;
+  g_autoptr(GFile) base = NULL;
+  g_autoptr(GFile) target = NULL;
+  struct statvfs sb;
+  gboolean ok;
+  guint64 tp;
+  guint64 ts;
+  int r;
+
+  /* Access time is not always supported, so support for it is best-effort.
+   */
+
+  r = statvfs (tt->path, &sb);
+  g_assert_cmpint (r, ==, 0);
+
+  if (bolt_flag_isset (sb.f_flag, ST_NOATIME))
+    {
+      g_test_skip ("file system mounted with noatime, can't modify access time");
+      return;
+    }
+
+  base = g_file_new_for_path (tt->path);
+  target = g_file_get_child (base, "this");
+
+  ok = g_file_query_exists (target, NULL);
+  g_assert_false (ok);
+
+  tp = 626648700;
+  ok = bolt_fs_touch (target, tp, 0, &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  info = g_file_query_info (target, TIME_QI, 0, NULL, &err);
+
+  g_assert_no_error (err);
+  g_assert_nonnull (info);
+
   ts = g_file_info_get_attribute_uint64 (info, "time::access");
   g_assert_cmpuint (ts, ==, tp);
 
-  ts = g_file_info_get_attribute_uint64 (info, "time::modified");
-  g_assert_cmpuint (ts, ==, 42);
-
-  /* omit mtime */
   ok = bolt_fs_touch (target, 42, 0, &err);
   g_assert_no_error (err);
   g_assert_true (ok);
@@ -1228,12 +1685,6 @@ test_fs_touch (TestIO *tt, gconstpointer user_data)
 
   ts = g_file_info_get_attribute_uint64 (info, "time::access");
   g_assert_cmpuint (ts, ==, 42);
-
-  /* mtime 0 means ignore, effectively meaning now, for touch */
-  ts = g_file_info_get_attribute_uint64 (info, "time::modified");
-  g_assert_cmpuint (ts, >=, now);
-  now = (guint64) g_get_real_time () / G_USEC_PER_SEC;
-  g_assert_cmpuint (ts, <=, now);
 }
 
 static void
@@ -1287,6 +1738,11 @@ test_str_erase (TestRng *tt, gconstpointer user_data)
   size_t n;
 
   bolt_get_random_data (buf, sizeof (buf) - 1);
+  buf[0] = 'b'; /* make sure we never have an empty string */
+  buf[1] = 'o';
+  buf[2] = 'l';
+  buf[3] = 'l';
+
   d1 = g_strdup (buf);
   d2 = g_strdup (buf);
 
@@ -1317,43 +1773,19 @@ test_str_erase (TestRng *tt, gconstpointer user_data)
 static void
 test_str_parse_int (TestRng *tt, gconstpointer user_data)
 {
-  struct
-  {
-    const char *str;
-    gint        val;
-    gboolean    error;
-  } table[] = {
-    {"0",                                        0,                FALSE},
-    {"1",                                        1,                FALSE},
-    {"-1",                                      -1,                FALSE},
-#if __SIZEOF_INT__ == 4
-    {"2147483647",                      2147483647,                FALSE}, /* MAX_INT */
-    {"-2147483648",                    -2147483648,                FALSE}, /* MIN_INT */
-    {"2147483648",                               0,                TRUE},  /* MAX_INT + 1 */
-    {"-2147483649",                              0,                TRUE},  /* MIN_INT - 1 */
-#elif __SIZEOF_INT__ == 8
-    {"9223372036854775807",    9223372036854775807,                FALSE}, /* MAX_INT */
-    {"-9223372036854775808",  -9223372036854775808,                FALSE}, /* MIN_INT */
-    {"9223372036854775808",                      0,                TRUE},  /* MAX_INT + 1 */
-    {"-9223372036854775809",                     0,                TRUE},  /* MIN_INT - 1 */
-#else
-    #warning __SIZEOF_INT__ not handled
-#endif
-    {"notanint",                                 0,                TRUE},
-    {"9223372036854775808",                      0,                TRUE}, /* overflow */
-    {"-9223372036854775809",                     0,                TRUE}, /* underflow */
-  };
-
-  for (gsize i = 0; i < G_N_ELEMENTS (table); i++)
+  for (gsize i = 0; i < G_N_ELEMENTS (str_to_int_table); i++)
     {
       g_autoptr(GError) error = NULL;
+      const char *txt = str_to_int_table[i].str;
+      gboolean expect_error = str_to_int_table[i].error;
+      gint val = str_to_int_table[i].val;
       gboolean ok;
       gint v;
 
       errno = 0;
-      ok = bolt_str_parse_as_int (table[i].str, &v, &error);
+      ok = bolt_str_parse_as_int (txt, &v, &error);
 
-      if (table[i].error)
+      if (expect_error)
         {
           int err = errno;
 
@@ -1363,7 +1795,43 @@ test_str_parse_int (TestRng *tt, gconstpointer user_data)
         }
       else
         {
-          g_assert_cmpuint (table[i].val, ==, v);
+          g_assert_cmpint (val, ==, v);
+          g_assert_true (ok);
+        }
+    }
+}
+
+static void
+test_str_parse_uint (TestRng *tt, gconstpointer user_data)
+{
+  for (gsize i = 0; i < G_N_ELEMENTS (str_to_uint_table); i++)
+    {
+      g_autoptr(GError) error = NULL;
+      const char *txt = str_to_uint_table[i].str;
+      gboolean expect_error = str_to_uint_table[i].error;
+      guint val = str_to_uint_table[i].val;
+      gboolean ok;
+      guint v;
+
+      errno = 0;
+      ok = bolt_str_parse_as_uint (txt, &v, &error);
+
+      if (g_test_verbose ())
+        g_test_message ("parsing '%s', expecting: %s", txt,
+                        (expect_error ? "error" : "success"));
+
+      if (expect_error)
+        {
+          int err = errno;
+
+          if (g_test_verbose ())
+            g_assert_nonnull (error);
+          g_assert_false (ok);
+          g_assert_cmpint (err, !=, 0);
+        }
+      else
+        {
+          g_assert_cmpuint (val, ==, v);
           g_assert_true (ok);
         }
     }
@@ -1394,8 +1862,50 @@ test_str_parse_uint64 (TestRng *tt, gconstpointer user_data)
       ok = bolt_str_parse_as_uint64 (table[i].str, &v, &err);
       if (table[i].error)
         {
+          int e = errno;
+
           g_assert_nonnull (err);
           g_assert_false (ok);
+          g_assert_cmpint (e, !=, 0);
+        }
+      else
+        {
+          g_assert_cmpuint (table[i].val, ==, v);
+          g_assert_true (ok);
+        }
+    }
+}
+
+static void
+test_str_parse_uint32 (TestRng *tt, gconstpointer user_data)
+{
+  struct
+  {
+    const char *str;
+    guint32     val;
+    gboolean    error;
+  } table[] = {
+    {"0",                    0,          FALSE},
+    {"1",                    1,          FALSE},
+    {"0xffffffff",           0xffffffff, FALSE}, /* G_MAXUINT64 */
+    {"notauint64",           0,          TRUE},
+    {"4294967296",           0,          TRUE}, /* overflow (G_MAXUINT32 + 1) */
+  };
+
+  for (gsize i = 0; i < G_N_ELEMENTS (table); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      gboolean ok;
+      guint32 v;
+
+      ok = bolt_str_parse_as_uint32 (table[i].str, &v, &err);
+      if (table[i].error)
+        {
+          int e = errno;
+
+          g_assert_nonnull (err);
+          g_assert_false (ok);
+          g_assert_cmpint (e, !=, 0);
         }
       else
         {
@@ -1467,6 +1977,25 @@ test_str_set (TestRng *tt, gconstpointer user_data)
   g_assert_cmpstr (target, ==, "Hallo Welt");
 }
 
+static void
+test_strv_make_n (TestRng *tt, gconstpointer user_data)
+{
+  g_auto(GStrv) empty = NULL;
+  g_auto(GStrv) full = NULL;
+  const char *check[] = {"voll", "voll", NULL};
+
+  empty = bolt_strv_make_n (0, "nichts");
+  g_assert_nonnull (empty);
+  g_assert_null (*empty);
+
+  full = bolt_strv_make_n (2, "voll");
+  g_assert_nonnull (full);
+  g_assert_nonnull (*full);
+  g_assert_cmpuint (bolt_gstrv_length0 (full), ==, 2);
+  bolt_assert_strv_equal (full, (GStrv) check, -1);
+}
+
+
 #define MAKE_GSTRV(...) (GStrv) (const char *[]){ __VA_ARGS__}
 
 static void
@@ -1486,6 +2015,7 @@ test_strv_length (TestRng *tt, gconstpointer user_data)
   for (gsize i = 0; i < G_N_ELEMENTS (table); i++)
     {
       g_assert_cmpuint (bolt_strv_length (table[i].strv), ==, table[i].l);
+      g_assert_cmpuint (bolt_gstrv_length0 (table[i].strv), ==, table[i].l);
 
       if (table[i].l == 0)
         g_assert_true (bolt_strv_isempty (table[i].strv));
@@ -1500,8 +2030,8 @@ test_strv_contains (TestRng *tt, gconstpointer user_data)
 
   const GStrv strv = MAKE_GSTRV ("a", "b", "c", "d", NULL);
 
-  g_assert_false (bolt_strv_contains (NULL, "nonexistant"));
-  g_assert_false (bolt_strv_contains (strv, "nonexistant"));
+  g_assert_false (bolt_strv_contains (NULL, "nonexistent"));
+  g_assert_false (bolt_strv_contains (strv, "nonexistent"));
 
   for (char **iter = strv; *iter; iter++)
     {
@@ -1733,14 +2263,66 @@ test_strv_rotate_left (TestRng *tt, gconstpointer user_data)
   g_assert_true (target == &a[4]);
 }
 
+#ifndef MAKE_GSTRV
+#define MAKE_GSTRV(...) (GStrv) (const char *[]){ __VA_ARGS__}
+#endif
+
+static void
+test_uuidv_check (TestRng *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  char *empty[] = {NULL};
+  const char *empty_entries[] = {"884c6edd-7118-4b21-b186-b02d396ecca0", "", NULL};
+  const char *valid[] = {"884c6edd-7118-4b21-b186-b02d396ecca0", NULL};
+  const GStrv invalid[] = {
+    MAKE_GSTRV ("\n", NULL),
+    MAKE_GSTRV ("884c6eddx7118x4b21xb186-b02d396ecca0", NULL),
+    MAKE_GSTRV ("884c6edd-4b21-b186-b02d396ecca0", NULL),
+    MAKE_GSTRV ("884c6edd-7118-4b21-b186-b02d396ecca0", "a", NULL),
+  };
+  gboolean ok;
+
+  ok = bolt_uuidv_check (NULL, TRUE, &err);
+  g_assert_true (ok);
+
+  ok = bolt_uuidv_check (empty, TRUE, &err);
+  g_assert_true (ok);
+
+  ok = bolt_uuidv_check ((GStrv) empty_entries, TRUE, &err);
+  g_assert_true (ok);
+
+  ok = bolt_uuidv_check (NULL, FALSE, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
+  g_assert_false (ok);
+  g_clear_error (&err);
+
+  ok = bolt_uuidv_check (empty, FALSE, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
+  g_assert_false (ok);
+  g_clear_error (&err);
+
+  ok = bolt_uuidv_check ((GStrv) empty_entries, FALSE, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
+  g_assert_false (ok);
+  g_clear_error (&err);
+
+  ok = bolt_uuidv_check ((GStrv) valid, TRUE, &err);
+  g_assert_true (ok);
+
+  for (gsize i = 0; i < G_N_ELEMENTS (invalid); i++)
+    {
+      ok = bolt_uuidv_check (invalid[i], FALSE, &err);
+      g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
+      g_assert_false (ok);
+      g_clear_error (&err);
+    }
+}
+
 static void
 test_term_fancy (TestRng *tt, gconstpointer user_data)
 {
-  if (bolt_is_fancy_terminal () == 0)
-    {
-      g_test_skip ("Terminal is not fancy");
-      return;
-    }
+  skip_test_unless (bolt_is_fancy_terminal (),
+                    "Terminal is not fancy");
 
   g_assert_cmpstr (bolt_color (ANSI_NORMAL), !=, "");
   g_assert_cmpstr (bolt_glyph (WARNING_SIGN), !=, "");
@@ -1749,11 +2331,8 @@ test_term_fancy (TestRng *tt, gconstpointer user_data)
 static void
 test_term_plain (TestRng *tt, gconstpointer user_data)
 {
-  if (bolt_is_fancy_terminal () != 0)
-    {
-      g_test_skip ("Terminal is too fancy");
-      return;
-    }
+  skip_test_if (bolt_is_fancy_terminal (),
+                "Terminal is too fancy");
 
   g_assert_cmpstr (bolt_color (ANSI_NORMAL), ==, "");
   g_assert_cmpstr (bolt_glyph (WARNING_SIGN), !=, "");
@@ -1812,17 +2391,22 @@ test_list_nh (TestRng *tt, gconstpointer user_data)
       g_assert_true (l[i].prev->next == &l[i]);
     }
 
-  c = 0;
-  bolt_nhlist_iter_init (&iter, l);
-  while ((k = bolt_nhlist_iter_next (&iter)))
+  /* start in the middle */
+  for (guint i = 0; i < 10; i++)
     {
-      BoltList *p = bolt_nhlist_iter_node (&iter);
-      g_assert_true (k == n + (c % 10));
-      g_assert_true (k == p);
-      c++;
-    }
+      c = 0;
+      bolt_nhlist_iter_init (&iter, &n[i]);
+      while ((k = bolt_nhlist_iter_next (&iter)))
+        {
+          BoltList *p = bolt_nhlist_iter_node (&iter);
+          g_assert_true (k == n + ((c + i) % 10));
+          g_assert_true (k == p);
+          c++;
+        }
 
-  g_assert_cmpuint (c, ==, 10);
+      g_assert_cmpuint (c, ==, 10);
+      g_debug ("start[%u] %p: count: %u", i, &n[i], c);
+    }
 }
 
 static void
@@ -1931,6 +2515,13 @@ main (int argc, char **argv)
               test_io_errors,
               test_io_tear_down);
 
+  g_test_add ("/common/io/mkdirat",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_mkdirat,
+              test_io_tear_down);
+
   g_test_add ("/common/io/verify",
               TestIO,
               NULL,
@@ -1938,6 +2529,41 @@ main (int argc, char **argv)
               test_io_verify,
               test_io_tear_down);
 
+  g_test_add ("/common/io/write_file_at",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_write_file_at,
+              test_io_tear_down);
+
+  g_test_add ("/common/io/write_int_at",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_write_int_at,
+              test_io_tear_down);
+
+  g_test_add ("/common/io/read_int_at",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_read_int_at,
+              test_io_tear_down);
+
+  g_test_add ("/common/io/write_uint_at",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_write_uint_at,
+              test_io_tear_down);
+
+  g_test_add ("/common/io/read_uint_at",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_read_uint_at,
+              test_io_tear_down);
+
   g_test_add ("/common/io/file_write_all",
               TestIO,
               NULL,
@@ -1945,6 +2571,13 @@ main (int argc, char **argv)
               test_io_file_write_all,
               test_io_tear_down);
 
+  g_test_add ("/common/io/renameat",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_renameat,
+              test_io_tear_down);
+
   g_test_add ("/common/io/copy_bytes",
               TestIO,
               NULL,
@@ -1952,6 +2585,13 @@ main (int argc, char **argv)
               test_io_copy_bytes,
               test_io_tear_down);
 
+  g_test_add ("/common/io/dir_is_empty",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_io_dir_is_empty,
+              test_io_tear_down);
+
   g_test_add ("/common/io/autoclose",
               TestIO,
               NULL,
@@ -1966,11 +2606,18 @@ main (int argc, char **argv)
               test_fs,
               test_io_tear_down);
 
-  g_test_add ("/common/fs/touch",
+  g_test_add ("/common/fs/touch/mtime",
               TestIO,
               NULL,
               test_io_setup,
-              test_fs_touch,
+              test_fs_touch_mtime,
+              test_io_tear_down);
+
+  g_test_add ("/common/fs/touch/atime",
+              TestIO,
+              NULL,
+              test_io_setup,
+              test_fs_touch_atime,
               test_io_tear_down);
 
   g_test_add ("/common/str",
@@ -1994,6 +2641,13 @@ main (int argc, char **argv)
               test_str_parse_int,
               NULL);
 
+  g_test_add ("/common/str/parse/uint",
+              TestRng,
+              NULL,
+              NULL,
+              test_str_parse_uint,
+              NULL);
+
   g_test_add ("/common/str/parse/uint64",
               TestRng,
               NULL,
@@ -2001,6 +2655,13 @@ main (int argc, char **argv)
               test_str_parse_uint64,
               NULL);
 
+  g_test_add ("/common/str/parse/uint32",
+              TestRng,
+              NULL,
+              NULL,
+              test_str_parse_uint32,
+              NULL);
+
   g_test_add ("/common/str/parse/boolean",
               TestRng,
               NULL,
@@ -2015,6 +2676,13 @@ main (int argc, char **argv)
               test_str_set,
               NULL);
 
+  g_test_add ("/common/strv/make_n",
+              TestRng,
+              NULL,
+              NULL,
+              test_strv_make_n,
+              NULL);
+
   g_test_add ("/common/strv/equal",
               TestRng,
               NULL,
@@ -2029,6 +2697,8 @@ main (int argc, char **argv)
               test_strv_length,
               NULL);
 
+
+
   g_test_add ("/common/strv/contains",
               TestRng,
               NULL,
@@ -2057,6 +2727,13 @@ main (int argc, char **argv)
               test_strv_rotate_left,
               NULL);
 
+  g_test_add ("/common/uuidv/check",
+              TestRng,
+              NULL,
+              NULL,
+              test_uuidv_check,
+              NULL);
+
   g_test_add ("/common/term/fancy",
               TestRng,
               NULL,
